r"""Packages
============
"""

import os
import re
from collections.abc import Iterator
from threading import Thread

from aiohttp import ClientSession, ClientTimeout
from jinja2 import Template
from pip._internal.metadata import get_default_environment
from pip._internal.metadata.base import BaseDistribution
from pip._vendor.packaging.utils import NormalizedName, canonicalize_name
from platformdirs import user_config_dir

pkginfos: dict[NormalizedName, str] = {}
NOT_FOUND = "Not found installed package!"
PATH = os.path.join(user_config_dir("pip"), "template.md.j2")
if not os.path.exists(PATH):
    PATH = os.path.join(
        os.path.join(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "assets"),
            "jinja2",
        ),
        "template.md.j2",
    )
with open(PATH) as f:
    TEMPLATE = f.read()
installed: dict[NormalizedName, BaseDistribution] = {
    dist.canonical_name: dist
    for dist in get_default_environment().iter_all_distributions()
}


def get_pkginfos() -> dict[NormalizedName, str]:
    r"""Get pkginfos.

    :rtype: dict[NormalizedName, str]
    """
    return pkginfos


def get_requiring_packages(current_dist: BaseDistribution) -> Iterator[str]:
    r"""Get requiring packages.

    :param current_dist:
    :type current_dist: BaseDistribution
    :rtype: Iterator[str]
    """
    return (
        dist.metadata["Name"] or "UNKNOWN"
        for dist in installed.values()
        if current_dist.canonical_name
        in {canonicalize_name(d.name) for d in dist.iter_dependencies()}
    )


async def update_pkginfos(timeout: int = 3) -> None:
    r"""Update pkginfos.

    :param timeout:
    :type timeout: int
    :rtype: None
    """
    global pkginfos
    for pkgname in installed:
        Thread(target=render_document, args=(pkgname, False, False)).start()
    await update_pkgnames(timeout)


async def update_pkgnames(timeout: int = 3):
    r"""Update pkgnames. IO bound.

    `<https://stackoverflow.com/questions/21419009/json-api-for-pypi-how-to-list-packages/51420285#51420285>`_

    :param timeout:
    :type timeout: int
    """
    global pkginfos
    try:
        async with (
            ClientSession() as session,
            session.get(
                "https://pypi.org/simple", timeout=ClientTimeout(total=timeout)
            ) as resp,
        ):
            text = await resp.text()
    except TimeoutError:
        return
    for match in re.finditer(r'"/simple/([^/]+)/"', text):
        pkgname: NormalizedName = match.groups()[0]  # type: ignore
        if pkgname not in pkginfos:
            pkginfos[pkgname] = NOT_FOUND


def render_document(
    pkgname: NormalizedName,
    has_requires: bool = True,
    has_required_by: bool = True,
):
    r"""Render document. CPU bound.

    :param pkgname:
    :type pkgname: NormalizedName
    :param has_requires:
    :type has_requires: bool
    :param has_required_by:
    :type has_required_by: bool
    """
    global pkginfos
    pkginfo = pkginfos.get(pkgname)
    # cache hit
    if pkginfo and not (pkginfo.find("1. ...") and pkginfo.find("2. ...")):
        return pkginfo
    if dist := installed.get(pkgname):
        requires = required_by = None
        if has_requires:
            requires = sorted(
                (req.name for req in dist.iter_dependencies()), key=str.lower
            )
        if has_required_by:
            required_by = sorted(get_requiring_packages(dist), key=str.lower)
        pkginfos[pkgname] = Template(TEMPLATE).render(
            dist=dist, requires=requires, required_by=required_by
        )
        return pkginfos[pkgname]
    return NOT_FOUND



r"""Utils
=========
"""

import os
from typing import Literal

from tree_sitter import Query
from tree_sitter_requirements._core import _language

from . import FILETYPE

QUERIES = {}


def get_query(name: str) -> Query:
    r"""Get query.

    :param name:
    :type name: str
    :rtype: Query
    """
    if name not in QUERIES:
        with open(
            os.path.join(
                os.path.join(
                    os.path.join(os.path.dirname(__file__), "assets"),
                    "queries",
                ),
                f"{name}{os.path.extsep}scm",
            )
        ) as f:
            text = f.read()
        QUERIES[name] = _language.query(text)
    return QUERIES[name]


def get_filetype(name: str) -> FILETYPE | Literal[""]:
    r"""Get filetype.

    :param name:
    :type name: str
    :rtype: FILETYPE | Literal[""]
    """
    return "pip"



r"""Test package"""

from requirements_language_server.packages import NOT_FOUND, render_document


class Test:
    r"""Test."""

    @staticmethod
    def test_get_document() -> None:
        r"""Test get document.

        :rtype: None
        """
        assert render_document("pip") != NOT_FOUND  # type: ignore



r"""Test utils."""

import os

from lsp_tree_sitter.diagnose import check
from requirements_language_server.finders import DIAGNOSTICS_FINDER_CLASSES
from requirements_language_server.utils import get_filetype
from tree_sitter_requirements import parse


class Test:
    r"""Test."""

    @staticmethod
    def test_check() -> None:
        r"""Test check.

        :rtype: None
        """
        result = check(
            [os.path.join(os.path.dirname(__file__), "requirements.txt.in")],
            parse,
            DIAGNOSTICS_FINDER_CLASSES,
            get_filetype,
        )
        assert result > 0



r"""This module can be called by
`python -m <https://docs.python.org/3/library/__main__.html>`_.
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from datetime import datetime

from . import __name__ as NAME
from . import __version__

try:
    import shtab
except ImportError:
    from . import _shtab as shtab

NAME = NAME.replace("_", "-")
VERSION = rf"""{NAME} {__version__}
Copyright (C) {datetime.now().year}
Written by Wu, Zhenyu
"""
EPILOG = """
Report bugs to <wuzhenyu@ustc.edu>.
"""


def get_parser() -> ArgumentParser:
    r"""Get a parser for unit test."""
    parser = ArgumentParser(
        epilog=EPILOG,
        formatter_class=RawDescriptionHelpFormatter,
    )
    shtab.add_argument_to(parser)
    parser.add_argument("--version", version=VERSION, action="version")
    parser.add_argument(
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="when to display color",
    )
    parser.add_argument(
        "--check",
        nargs="*",
        default={},
        help="check file's errors and warnings",
    ).complete = shtab.FILE  # type: ignore
    parser.add_argument(
        "--format",
        nargs="*",
        default={},
        help="format files",
    ).complete = shtab.FILE  # type: ignore
    return parser


def main() -> None:
    r"""Parse arguments and provide shell completions."""
    args = get_parser().parse_args()

    from lsp_tree_sitter.diagnose import check
    from lsp_tree_sitter.format import format
    from tree_sitter_requirements import parse

    from .finders import DIAGNOSTICS_FINDER_CLASSES, FORMATTING_FINDER_CLASSES
    from .utils import get_filetype

    if args.format or args.check:
        format(args.format, parse, FORMATTING_FINDER_CLASSES, get_filetype)  # type: ignore
        exit(
            check(
                args.check,
                parse,
                DIAGNOSTICS_FINDER_CLASSES,
                get_filetype,
                args.color,
            )
        )

    from .server import RequirementsLanguageServer

    RequirementsLanguageServer(NAME, __version__).start_io()


if __name__ == "__main__":
    main()



r"""Finders
===========
"""

import os
from dataclasses import dataclass
from typing import Any, Literal

import tree_sitter_requirements as requirements
from lsp_tree_sitter import UNI
from lsp_tree_sitter.finders import (
    ErrorFinder,
    QueryFinder,
    RepeatedFinder,
    UnsortedFinder,
)
from lsprotocol.types import Diagnostic, DiagnosticSeverity
from tree_sitter import Node, Tree

from . import FILETYPE
from .packages import get_pkginfos
from .utils import get_query


@dataclass(init=False)
class InvalidPackageFinder(QueryFinder):
    r"""Invalidpackagefinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such package",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        query = get_query("package")
        super().__init__(query, message, severity)

    def capture2uni(self, capture: tuple[Node, str], uri: str) -> UNI | None:
        r"""Capture2uni.

        :param capture:
        :type capture: tuple[Node, str]
        :param uri:
        :type uri: str
        :rtype: UNI | None
        """
        node, _ = capture
        uni = UNI(uri, node)
        text = uni.get_text()
        return uni if text not in get_pkginfos() else None


@dataclass(init=False)
class OptionFinder(QueryFinder):
    r"""Optionfinder."""

    def __init__(
        self,
        filetype: FILETYPE,
        message: str = "setuptools only supports PEP508",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param filetype:
        :type filetype: FILETYPE
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        query = get_query("option")
        super().__init__(query, message, severity)
        if filetype == "pip":
            self.find_all = self.fake_find_all

    def fake_find_all(self, *args: Any, **kwargs: Any) -> list:
        r"""Fake find all.

        :param args:
        :type args: Any
        :param kwargs:
        :type kwargs: Any
        :rtype: list
        """
        return []


@dataclass(init=False)
class InvalidPathFinder(QueryFinder):
    r"""Invalidpathfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such {{_type}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        query = get_query("path")
        super().__init__(query, message, severity)
        self.paths = {"file": [], "dir": []}

    def captures2unis(
        self, captures: list[tuple[Node, str]], uri: str
    ) -> list[UNI]:
        r"""Captures2unis.

        :param captures:
        :type captures: list[tuple[Node, str]]
        :param uri:
        :type uri: str
        :rtype: list[UNI]
        """
        unis = []
        for capture in captures:
            node, label = capture
            uni = UNI(uri, node)
            path = uni.get_path()
            if (
                label == "file"
                and not os.path.isfile(path)
                or label == "dir"
                and not os.path.isdir(path)
            ):
                unis += [uni]
                self.paths[label] += [uni]
        return unis

    def get_type(self, uni: UNI) -> Literal["file", "dir"]:
        try:
            self.paths["file"].index(uni)
            return "file"
        except ValueError:
            return "dir"

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        return uni.get_diagnostic(
            self.message,
            self.severity,
            _type=self.get_type(uni),
        )


class RepeatedPackageFinder(RepeatedFinder):
    r"""Repeatedpackagefinder."""

    def __init__(
        self,
        filetype: FILETYPE = "pip",
        message: str = "{{uni.get_text()}}: is repeated on {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param self:
        :param pep508:
        :type pep508: bool
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.filetype = filetype

    def is_include_node(self, node: Node) -> bool:
        r"""Is include node.

        :param node:
        :type node: Node
        :rtype: bool
        """
        return node.type == "path" and self.filetype == "pip"

    def parse(self, code: bytes) -> Tree:
        r"""Parse.

        :param code:
        :type code: bytes
        :rtype: Tree
        """
        return requirements.parse(code)

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return uni.node.type == "package"


class UnsortedRequirementFinder(UnsortedFinder):
    r"""Unsortedrequirementfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: is unsorted due to {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return uni.node.type == "requirement"


FORMATTING_FINDER_CLASSES = [UnsortedRequirementFinder]


DIAGNOSTICS_FINDER_CLASSES = [
    ErrorFinder,
    InvalidPackageFinder,
    InvalidPathFinder,
    RepeatedPackageFinder,
    OptionFinder,
    UnsortedRequirementFinder,
]
